#create a rectangle latitude/longitude based mesh
#we will use some GIS way to define it
#longitude left and latitude bottom and nrow and ncolumn and resolution is used to define the rectangle
#because it is mesh, it represent the edge instead of center
#we will use gdal api for most operations
import os, sys
import numpy as np
from osgeo import ogr, osr
from pyflowline.classes.latlon import pylatlon
from pyflowline.formats.convert_coordinates import convert_gcs_coordinates_to_cell

def create_latlon_mesh(dLongitude_left_in, dLatitude_bot_in, dResolution_degree_in, ncolumn_in, nrow_in, sFilename_output_in):

   
    if os.path.exists(sFilename_output_in): 
        #delete it if it exists
        os.remove(sFilename_output_in)


    #pDriver_shapefile = ogr.GetDriverByName('Esri Shapefile')
    pDriver_geojson = ogr.GetDriverByName('GeoJSON')
    
    pDataset = pDriver_geojson.CreateDataSource(sFilename_output_in)
    pSpatial_reference_gcs = osr.SpatialReference()  
    pSpatial_reference_gcs.ImportFromEPSG(4326)    # WGS84 lat/lon
    pSpatial_reference_gcs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

    pLayer = pDataset.CreateLayer('cell', pSpatial_reference_gcs, ogr.wkbPolygon)
    # Add one attribute
    pLayer.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
    pLayer.CreateField(ogr.FieldDefn('lon', ogr.OFTReal)) #long type for high resolution
    pLayer.CreateField(ogr.FieldDefn('lat', ogr.OFTReal)) #long type for high resolution
    pArea_field = ogr.FieldDefn('area', ogr.OFTReal)
    pArea_field.SetWidth(20)
    pArea_field.SetPrecision(2)
    pLayer.CreateField(pArea_field)
    
    pLayerDefn = pLayer.GetLayerDefn()
    pFeature = ogr.Feature(pLayerDefn)

    

    xleft = dLongitude_left_in
    xspacing= dResolution_degree_in
    ybottom = dLatitude_bot_in
    yspacing = dResolution_degree_in

    lID =0 
    #.........
    #(x2,y2)-----(x3,y3)
    #   |           |
    #(x1,y1)-----(x4,y4)
    #...............
    aLatlon = list()
    for column in range(0, ncolumn_in):
        for row in range(0, nrow_in):
            #define a polygon here
            x1 = xleft + (column * xspacing)
            y1 = ybottom + (row * yspacing)

            x2 = xleft + (column * xspacing)
            y2 = ybottom + ((row + 1) * yspacing)

            x3 = xleft + ((column + 1) * xspacing)
            y3 = ybottom + ((row + 1) * yspacing)

            x4 = xleft + ((column + 1) * xspacing)
            y4 = ybottom + (row * yspacing)
           

            ring = ogr.Geometry(ogr.wkbLinearRing)
            ring.AddPoint(x1, y1)
            ring.AddPoint(x2, y2)
            ring.AddPoint(x3, y3)
            ring.AddPoint(x4, y4)
            ring.AddPoint(x1, y1)
            pPolygon = ogr.Geometry(ogr.wkbPolygon)
            pPolygon.AddGeometry(ring)

            dLon = (x1 + x2 + x3 + x4)/4.0
            dLat = (y1 + y2 + y3 + y4)/4.0
            aCoords = np.full((5,2), -9999.0, dtype=float)
            aCoords[0,0] = x1
            aCoords[0,1] = y1
            aCoords[1,0] = x2
            aCoords[1,1] = y2
            aCoords[2,0] = x3
            aCoords[2,1] = y3
            aCoords[3,0] = x4
            aCoords[3,1] = y4
            aCoords[4,0] = x1
            aCoords[4,1] = y1
            dummy1= np.array(aCoords)
            

            pLatlon = convert_gcs_coordinates_to_cell(3, dLon, dLat, dummy1)
            dArea = pLatlon.calculate_cell_area()


            pFeature.SetGeometry(pPolygon)
            pFeature.SetField("id", lID)
            pFeature.SetField("lon", dLon )
            pFeature.SetField("lat", dLat )
            pFeature.SetField("area", dArea )
            pLayer.CreateFeature(pFeature)

            lID = lID + 1
            aLatlon.append(pLatlon)


            pass
        
    pDataset = pLayer = pFeature  = None      



    return aLatlon








import sys
from pathlib import Path
from os.path import realpath
import argparse
import logging
for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)

logging.basicConfig(format='%(asctime)s %(message)s')
logging.warning('is the time Pyflowline simulation started.')
def is_module_available(module_name):
    if sys.version_info < (3, 0):
        # python 2
        import importlib
        torch_loader = importlib.find_loader(module_name)
    elif sys.version_info <= (3, 3):
        # python 3.0 to 3.3
        import pkgutil
        torch_loader = pkgutil.find_loader(module_name)
    elif sys.version_info >= (3, 4):
        # python 3.4 and above
        import importlib
        torch_loader = importlib.util.find_spec(module_name)

    return torch_loader is not None

iFlag = is_module_available('pyflowline')
import pyflowline
from pyflowline.classes.pycase import flowlinecase
from pyflowline.pyflowline_read_model_configuration_file import pyflowline_read_model_configuration_file
from pyflowline.pyflowline_generate_template_configuration_json_file import pyflowline_generate_template_configuration_json_file

parser = argparse.ArgumentParser()
parser.add_argument("--sMesh_type", help = "sMesh_type",  type = str)
parser.add_argument("--iCase_index", help = "iCase_index",  type = int)
parser.add_argument("--dResolution_meter", help = "dResolution_meter",  type = float)
parser.add_argument("--sDate", help = "sDate",  type = str)

#example
#python notebook.py --sMesh_type hexagon --iCase_index 1 --dResolution_meter 50000 --sDate 20220201
pArgs = parser.parse_args()
if len(sys.argv) == 1:
    sMesh_type = 'mpas'
    iCase_index = 1
    dResolution_meter=5000
    sDate='20220308'
else:
    if len(sys.argv)> 1:
        sMesh_type = pArgs.sMesh_type
        iCase_index = pArgs.iCase_index
        dResolution_meter=pArgs.dResolution_meter
        sDate = pArgs.sDate
        print(sMesh_type, iCase_index, dResolution_meter, sDate)
    else:
        print(len(sys.argv), 'Missing arguaments')
        pass

sPath = str( Path().resolve() )
iFlag_option = 1
sWorkspace_data = realpath( sPath +  '/data/susquehanna' )
sWorkspace_input =  str(Path(sWorkspace_data)  /  'input')
sWorkspace_output=  str(Path(sWorkspace_data)  /  'output')
if iFlag_option == 1:

    
    sFilename_configuration_in = realpath( sPath +  '/tests/configurations/template.json' )
    
    oPyflowline = pyflowline_generate_template_configuration_json_file(sFilename_configuration_in,\
         sWorkspace_input, sWorkspace_output, sMesh_type_in=sMesh_type, iCase_index_in = iCase_index, sDate_in = sDate)
    
    
    print(oPyflowline.tojson())
else: 
    if iFlag_option == 2:
        #an example configuration file is provided with the repository, but you need to update this file based on your own case study        
        if sMesh_type=='hexagon':
            sFilename_configuration_in = realpath( sPath +  '/tests/configurations/pyflowline_susquehanna_hexagon.json' )
        else:
            if sMesh_type=='square':
                sFilename_configuration_in = realpath( sPath +  '/test/configurations/pyflowline_susquehanna_square.json' )
            else:
                if sMesh_type=='latlon':
                    sFilename_configuration_in = realpath( sPath +  '/tests/configurations/pyflowline_susquehanna_latlon.json' )
                else:
                    sFilename_configuration_in = realpath( sPath +  '/tests/configurations/)pyflowline_susquehanna_mpas.json' )
        
        
        oPyflowline = pyflowline_read_model_configuration_file(sFilename_configuration_in, iCase_index_in=iCase_index, dResolution_meter_in=dResolution_meter, sDate_in=sDate)
        

#pyflowline can process multiple basins within one singel run
#the total number of basin is controlled by the nOutlet variable
#convert the raw flowline into geojson in WGS84 system        
oPyflowline.convert_flowline_to_json()

oPyflowline.flowline_simplification()

oPyflowline.mesh_generation()

oPyflowline.reconstruct_topological_relationship()

oPyflowline.analyze()

oPyflowline.export()



print('Finished')

logging.basicConfig(format='%(asctime)s %(message)s')
logging.warning('is the time Pyflowline simulation finished.')



import os
from pathlib import Path
from abc import ABCMeta
import json
from json import JSONEncoder
import datetime

import numpy as np

from osgeo import ogr, osr



from shapely.wkt import loads


from pyflowline.classes.mpas import pympas
from pyflowline.classes.hexagon import pyhexagon
from pyflowline.classes.latlon import pylatlon
from pyflowline.classes.square import pysquare
from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.basin import pybasin
from pyflowline.classes.flowline import pyflowline
from pyflowline.classes.edge import pyedge

from pyflowline.formats.convert_shapefile_to_json import convert_shapefile_to_json_swat
from pyflowline.formats.read_mesh import read_mesh_json

from pyflowline.algorithms.auxiliary.text_reader_string import text_reader_string
from pyflowline.algorithms.auxiliary.gdal_functions import reproject_coordinates, reproject_coordinates_batch
from pyflowline.algorithms.auxiliary.gdal_functions  import degree_to_meter
from pyflowline.algorithms.auxiliary.gdal_functions  import meter_to_degree
from pyflowline.algorithms.auxiliary.gdal_functions import gdal_read_geotiff_file, retrieve_geotiff_metadata
from pyflowline.algorithms.auxiliary.gdal_functions import  Google_MetersPerPixel

from pyflowline.mesh.hexagon.create_hexagon_mesh import create_hexagon_mesh
from pyflowline.mesh.latlon.create_latlon_mesh import create_latlon_mesh
from pyflowline.mesh.square.create_square_mesh import create_square_mesh
from pyflowline.mesh.mpas.create_mpas_mesh import create_mpas_mesh
from pyflowline.mesh.tin.create_tin_mesh import create_tin_mesh



pDate = datetime.datetime.today()
sDate_default = "{:04d}".format(pDate.year) + "{:02d}".format(pDate.month) + "{:02d}".format(pDate.day)

class CaseClassEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.float32):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, list):
            pass  
        if isinstance(obj, pyvertex):
            return json.loads(obj.tojson()) #lVertexID
        if isinstance(obj, pyedge):
            return obj.lEdgeID        
        if isinstance(obj, pyflowline):
            return obj.lFlowlineID
        if isinstance(obj, pyhexagon):
            return obj.lCellID
        if isinstance(obj, pysquare):
            return obj.lCellID
        if isinstance(obj, pylatlon):
            return obj.lCellID
        if isinstance(obj, pympas):
            return obj.lCellID        
        if isinstance(obj, pybasin):
            return obj.lBasinID    
            
        return JSONEncoder.default(self, obj)

class flowlinecase(object):
    __metaclass__ = ABCMeta
    iCase_index= 0
    sMesh_type = 1
    iFlag_standalone=1
    iFlag_global = 0
    iFlag_multiple_outlet = 0
    iFlag_use_shapefile_extent=1
    iFlag_use_mesh_dem=0
    iFlag_save_mesh = 0
    iFlag_simplification = 1 #user can turn on/off
    iFlag_create_mesh=1
    iFlag_intersect = 1
    
    iFlag_rotation=0

    nOutlet = 1 #by default , there shoule ne only one ouelet
    dResolution_degree=0.0
    dResolution_meter=0.0
    dThreshold_small_river=0.0
    dLongitude_left = -180
    dLongitude_right = 180
    dLatitude_bot = -90
    dLatitude_top = 90
    sFilename_model_configuration=''
    sWorkspace_input=''      
    sWorkspace_output=''    
    #sWorkspace_output_case=''    
    sRegion=''
    sModel=''
    sMesh_type ='hexagon'

    sCase=''
    sDate=''    

    sFilename_spatial_reference=''
    sFilename_dem=''   

    sFilename_mesh=''
    sFilename_mesh_info=''
    sFilename_mesh_netcdf=''
    

    aBasin = list()
    aFlowline_simplified=list()
    aFlowline_conceptual=list()
    aCellID_outlet = list()
    aCell=list()

    
    def __init__(self, aConfig_in,\
            iFlag_standalone_in= None,\
            sModel_in = None,\
            sWorkspace_output_in = None):
        #flags
        if iFlag_standalone_in is not None:
            self.iFlag_standalone = iFlag_standalone_in
        else:
            if 'iFlag_standalone' in aConfig_in:
                self.iFlag_standalone = int(aConfig_in['iFlag_standalone'])
            else:
                self.iFlag_standalone=1

        if 'iFlag_flowline' in aConfig_in:
            self.iFlag_flowline             = int(aConfig_in[ 'iFlag_flowline'])
        
        if 'iFlag_global' in aConfig_in:
            self.iFlag_global             = int(aConfig_in[ 'iFlag_global'])
        
        if 'iFlag_multiple_outlet' in aConfig_in:
            self.iFlag_multiple_outlet             = int(aConfig_in[ 'iFlag_multiple_outlet'])  
        
        if 'iFlag_simplification' in aConfig_in:
            self.iFlag_simplification = int(aConfig_in['iFlag_simplification'])


        if 'iFlag_create_mesh' in aConfig_in:
            self.iFlag_create_mesh = int(aConfig_in['iFlag_create_mesh'])    
        
        if 'iFlag_save_mesh' in aConfig_in:
            self.iFlag_save_mesh             = int(aConfig_in[ 'iFlag_save_mesh'])
        
        if 'iFlag_use_mesh_dem' in aConfig_in:
            self.iFlag_use_mesh_dem = int(aConfig_in['iFlag_use_mesh_dem'])

        if 'iFlag_use_shapefile_extent' in aConfig_in:
            self.iFlag_use_shapefile_extent = int(aConfig_in['iFlag_use_shapefile_extent'])    

        if 'iFlag_rotation' in aConfig_in:
            self.iFlag_rotation = int(aConfig_in['iFlag_rotation'])
        

        if 'iFlag_intersect' in aConfig_in:
            self.iFlag_intersect = int(aConfig_in['iFlag_intersect'])

        if 'nOutlet' in aConfig_in:
            self.nOutlet = int(aConfig_in['nOutlet'])
             
        if 'iCase_index' in aConfig_in:
            iCase_index = int(aConfig_in['iCase_index'])
        else:
            iCase_index = 1
        sCase_index = "{:03d}".format( iCase_index )
        self.iCase_index =   iCase_index

        if 'dResolution_degree' in aConfig_in:
            self.dResolution_degree = float(aConfig_in['dResolution_degree']) 

        if 'dResolution_meter' in aConfig_in:
            self.dResolution_meter = float(aConfig_in['dResolution_meter']) 
        else:
            print('Please specify resolution.')

        if 'dLongitude_left' in aConfig_in:
            self.dLongitude_left = float(aConfig_in['dLongitude_left']) 

        if 'dLongitude_right' in aConfig_in:
            self.dLongitude_right = float(aConfig_in['dLongitude_right']) 

        if 'dLatitude_bot' in aConfig_in:
            self.dLatitude_bot = float(aConfig_in['dLatitude_bot']) 

        if 'dLatitude_top' in aConfig_in:
            self.dLatitude_top = float(aConfig_in['dLatitude_top']) 

        if 'sFilename_model_configuration' in aConfig_in:
            self.sFilename_model_configuration    = aConfig_in[ 'sFilename_model_configuration']

        if 'sFilename_spatial_reference' in aConfig_in:
            self.sFilename_spatial_reference = aConfig_in['sFilename_spatial_reference']

        if 'sFilename_dem' in aConfig_in:
            self.sFilename_dem = aConfig_in['sFilename_dem']

        if 'sFilename_mesh_netcdf' in aConfig_in:
            self.sFilename_mesh_netcdf = aConfig_in['sFilename_mesh_netcdf']
        
        if 'sWorkspace_bin' in aConfig_in:
            self.sWorkspace_bin= aConfig_in[ 'sWorkspace_bin']
            
        if 'sWorkspace_input' in aConfig_in:
            self.sWorkspace_input = aConfig_in[ 'sWorkspace_input']
        
       
        
        if sWorkspace_output_in is not None:
            self.sWorkspace_output = sWorkspace_output_in
        else:
            if 'sWorkspace_output' in aConfig_in:
                self.sWorkspace_output = aConfig_in[ 'sWorkspace_output']
        
        if 'sJob' in aConfig_in:
            self.sJob =  aConfig_in['sJob'] 
        if 'sRegion' in aConfig_in:
            self.sRegion               = aConfig_in[ 'sRegion']
        
        if sModel_in is not None:
            self.sModel = sModel_in
        else:
            if 'sModel' in aConfig_in:
                self.sModel                = aConfig_in[ 'sModel']

                      
        sDate   = aConfig_in[ 'sDate']
        if sDate is not None:
            self.sDate= sDate
        else:
            self.sDate = sDate_default
        
        
        sCase = self.sModel  + self.sDate + sCase_index
        self.sCase = sCase

        if 'sMesh_type' in aConfig_in:
            self.sMesh_type =  aConfig_in['sMesh_type']
        else:
            self.sMesh_type = 'hexagon'

                
        
        sMesh_type = self.sMesh_type
        if sMesh_type =='hexagon': #hexagon
            self.iMesh_type = 1
        else:
            if sMesh_type =='square': #square
                self.iMesh_type = 2
            else:
                if sMesh_type =='latlon': #latlon
                    self.iMesh_type = 3
                else:
                    if sMesh_type =='mpas': #mpas
                        self.iMesh_type = 4
                    else:
                        if sMesh_type =='tin': #tin
                            self.iMesh_type = 5
                        else:
                            print('Unsupported mesh type?')
        
        #the model can be run as part of hexwatershed or standalone
        if self.iFlag_standalone == 1:
            #in standalone case, will add case information and update output path
            sPath = str(Path(self.sWorkspace_output)  /  sCase)
            self.sWorkspace_output = sPath
        else:
            #use specified output path, also do not add output or input tag
            sPath = self.sWorkspace_output
        
        Path(sPath).mkdir(parents=True, exist_ok=True)

        self.aBasin = list()
        if self.iFlag_flowline == 1:
            if 'sFilename_basins' in aConfig_in:
                self.sFilename_basins = aConfig_in['sFilename_basins']
                with open(self.sFilename_basins) as json_file:
                    dummy_data = json.load(json_file)     
                    for i in range(self.nOutlet):
                        sBasin =  "{:03d}".format(i+1)   
                        dummy_basin = dummy_data[i]
                        dummy_basin['sWorkspace_output_basin'] = str(Path(self.sWorkspace_output) / sBasin )
                        
                        pBasin = pybasin(dummy_basin)
    
                        self.aBasin.append(pBasin)
            else:
                pass
     

        #model generated files   
      
        self.sFilename_mesh = os.path.join(str(Path(self.sWorkspace_output)  ) , sMesh_type + ".json" )               
        self.sFilename_mesh_info= os.path.join(str(Path(self.sWorkspace_output)  ) , sMesh_type + "_mesh_info.json"  )    
                
        return
        

    def convert_flowline_to_json(self):
        for pBasin in self.aBasin:            
            pBasin.convert_flowline_to_json()
            pass
 
    def flowline_simplification(self):
        aFlowline_out = list()   #store all the flowline
        if self.iFlag_simplification == 1: 
            for pBasin in self.aBasin:
                aFlowline_basin = pBasin.flowline_simplification()                
                aFlowline_out = aFlowline_out + aFlowline_basin

            self.aFlowline_simplified = aFlowline_out

        #export the outlet into a since file
        aOutlet = list()
        if self.iFlag_simplification == 1: 
            for pBasin in self.aBasin:
                aOutlet.append(pBasin.pVertex_outlet)

        return aFlowline_out
    
    def mesh_generation(self):        
        iFlag_global =  self.iFlag_global
        iMesh_type = self.iMesh_type
        iFlag_save_mesh = self.iFlag_save_mesh
        iFlag_rotation = self.iFlag_rotation
        dResolution_degree = self.dResolution_degree
        dResolution_meter = self.dResolution_meter
        sFilename_dem = self.sFilename_dem
        sFilename_spatial_reference = self.sFilename_spatial_reference
        sFilename_mesh = self.sFilename_mesh
        if iMesh_type !=4: #hexagon
            spatial_reference_target = osr.SpatialReference()  
            spatial_reference_target.ImportFromEPSG(4326)
            if self.iFlag_use_shapefile_extent==1:
                pDriver_shapefile = ogr.GetDriverByName('Esri Shapefile')
                pDataset_shapefile = pDriver_shapefile.Open(self.sFilename_spatial_reference, 0)
                pLayer_shapefile = pDataset_shapefile.GetLayer(0)
                pSpatial_reference = pLayer_shapefile.GetSpatialRef()
                (dOriginX, dX_right, dY_bot, dOriginY) = pLayer_shapefile.GetExtent() # not in same order as ogrinfo
                dLongitude_left,  dLatitude_bot= reproject_coordinates(dOriginX, dY_bot,pSpatial_reference,    spatial_reference_target)
                dLongitude_right, dLatitude_top= reproject_coordinates(dX_right, dOriginY,pSpatial_reference,  spatial_reference_target)
                dLatitude_mean = 0.5 * (dLatitude_top + dLatitude_bot)
                pass
            else:
                dPixelWidth, dOriginX, dOriginY, nrow, ncolumn, pSpatialRef_dem, pProjection, pGeotransform\
                     = retrieve_geotiff_metadata(sFilename_dem)

                dY_bot = dOriginY - (nrow+1) * dPixelWidth
                dLongitude_left,  dLatitude_bot= reproject_coordinates(dOriginX, dY_bot,pSpatialRef_dem,    spatial_reference_target)
                dX_right = dOriginX + (ncolumn +1) * dPixelWidth

                dLongitude_right, dLatitude_top= reproject_coordinates(dX_right, dOriginY,pSpatialRef_dem,  spatial_reference_target)
                dLatitude_mean = 0.5 * (dLatitude_top + dLatitude_bot)
                pass

            if dResolution_meter < 0:
                #not used
                pass
            else:
                dResolution_degree = meter_to_degree(dResolution_meter, dLatitude_mean)

            dX_left = dOriginX
            dY_top = dOriginY
        else:
            pass

        if iMesh_type ==1: #hexagon

            #hexagon edge
            dResolution_meter = degree_to_meter(dLatitude_mean, dResolution_degree )
            dArea = np.power(dResolution_meter,2.0)
            dLength_edge = np.sqrt(  2.0 * dArea / (3.0* np.sqrt(3.0))  )
            if iFlag_rotation ==0:            
                dX_spacing = dLength_edge * np.sqrt(3.0)
                dY_spacing = dLength_edge * 1.5
                ncolumn= int( (dX_right - dX_left) / dX_spacing )
                nrow= int( (dY_top - dY_bot) / dY_spacing ) 
            else:            
                dX_spacing = dLength_edge * 1.5
                dY_spacing = dLength_edge * np.sqrt(3.0)    
                ncolumn= int( (dX_right - dX_left) / dX_spacing )+1
                nrow= int( (dY_top - dY_bot) / dY_spacing )

            aHexagon = create_hexagon_mesh(iFlag_rotation, dX_left, dY_bot, dResolution_meter, ncolumn, nrow, \
                sFilename_mesh, sFilename_spatial_reference)
            return aHexagon
        else:
            if iMesh_type ==2: #sqaure
                ncolumn= int( (dX_right - dX_left) / dResolution_meter )
                nrow= int( (dY_top - dY_bot) / dResolution_meter )

                aSquare = create_square_mesh(dX_left, dY_bot, dResolution_meter, ncolumn, nrow, \
                    sFilename_mesh, sFilename_spatial_reference)
                return aSquare
            else:
                if iMesh_type ==3: #latlon
                    dResolution_meter = degree_to_meter(dLatitude_mean, dResolution_degree)
                    dArea = np.power(dResolution_meter,2.0)
                    dLatitude_top    = self.dLatitude_top   
                    dLatitude_bot    = self.dLatitude_bot   
                    dLongitude_left  = self.dLongitude_left 
                    dLongitude_right = self.dLongitude_right
                    ncolumn= int( (dLongitude_right - dLongitude_left) / dResolution_degree )
                    nrow= int( (dLatitude_top - dLatitude_bot) / dResolution_degree )
                    aLatlon = create_latlon_mesh(dLongitude_left, dLatitude_bot, dResolution_degree, ncolumn, nrow, \
                        sFilename_mesh)
                    return aLatlon
                else:
                    if iMesh_type == 4: #mpas
                        iFlag_use_mesh_dem = self.iFlag_use_mesh_dem
                        sFilename_mesh_netcdf = self.sFilename_mesh_netcdf
                        dLatitude_top    = self.dLatitude_top   
                        dLatitude_bot    = self.dLatitude_bot   
                        dLongitude_left  = self.dLongitude_left 
                        dLongitude_right = self.dLongitude_right
                        aMpas = create_mpas_mesh(iFlag_global, iFlag_use_mesh_dem, iFlag_save_mesh, \
                              dLongitude_left, dLongitude_right,  dLatitude_top, dLatitude_bot, \
                                    sFilename_mesh_netcdf,      sFilename_mesh)
                        return aMpas
                    else:
                        if iMesh_type ==5: #tin this one need to be updated because central location issue
                            #tin edge
                            dArea = np.power(dResolution_meter,2.0)
                            dLength_edge = np.sqrt(  4.0 * dArea /  np.sqrt(3.0) )  
                            dX_shift = 0.5 * dLength_edge
                            dY_shift = 0.5 * dLength_edge * np.sqrt(3.0) 
                            dX_spacing = dX_shift * 2
                            dY_spacing = dY_shift
                            ncolumn= int( (dX_right - dX_left) / dX_shift )
                            nrow= int( (dY_top - dY_bot) / dY_spacing ) 
                            aTin = create_tin_mesh(dX_left, dY_bot, dResolution_meter, ncolumn, nrow,sFilename_mesh, sFilename_spatial_reference)
                            return aTin
                        else:
                            print('Unsupported mesh type?')
                            return
        return
    
    def reconstruct_topological_relationship(self):
        iMesh_type = self.iMesh_type
        iFlag_intersect = self.iFlag_intersect
        sWorkspace_output = self.sWorkspace_output
        nOutlet = self.nOutlet
        sFilename_mesh=self.sFilename_mesh
        self.aCell, pSpatial_reference_mesh = read_mesh_json(iMesh_type, sFilename_mesh)
        
        aFlowline_conceptual = list()   #store all the flowline
        aCellID_outlet = list()
        aBasin = list()
        aCell_intersect=list()
        if iFlag_intersect == 1:
            for pBasin in self.aBasin:
                aCell_intersect_basin = pBasin.reconstruct_topological_relationship(iMesh_type,sFilename_mesh)
                aFlowline_conceptual = aFlowline_conceptual + pBasin.aFlowline_basin_conceptual
                aBasin.append(pBasin)
                aCellID_outlet.append(pBasin.lCellID_outlet)
                aCell_intersect = aCell_intersect + aCell_intersect_basin

                #update length?
            for pCell in self.aCell:
                for pCell2 in aCell_intersect:
                    if pCell2.lCellID == pCell.lCellID:
                        pCell.dLength_flowline = pCell2.dLength_flowline

            
            #save basin json info for hexwatershed model
            sPath = os.path.dirname(self.sFilename_basins)
            sName = str(Path(self.sFilename_basins).stem ) + '_new.json'
            sFilename_configuration  =  os.path.join( sPath  , sName)

            with open(sFilename_configuration, 'w', encoding='utf-8') as f:
                sJson = json.dumps([json.loads(ob.tojson()) for ob in aBasin],\
                    sort_keys=True, \
                    indent = 4)        

                f.write(sJson)    
                f.close()
            
            self.aFlowline_conceptual = aFlowline_conceptual
            self.aCellID_outlet = aCellID_outlet

            return   aFlowline_conceptual, aCellID_outlet

        return

    def analyze(self):
        for pBasin in self.aBasin:
            pBasin.analyze()
        return
        
    

    def setup(self):
        self.convert_flowline_to_json()
        return

    def run(self):
        self.flowline_simplification()
        self.mesh_generation()
        self.reconstruct_topological_relationship()
        self.export()
        return

   
    def export(self):
        
        self.export_mesh_info_to_json()
        for pBasin in self.aBasin:
            pBasin.export()

        self.tojson()

    def export_mesh_info_to_json(self):
        
        aCell_all = self.aCell
        sFilename_json = self.sFilename_mesh_info
        ncell=len(aCell_all)
        iFlag_flowline = self.iFlag_flowline
        aCellID_outlet = self.aCellID_outlet

        aFlowline = self.aFlowline_conceptual

        if iFlag_flowline == 1:
            nFlowline = len(aFlowline)
            for i in range(nFlowline):
                pFlowline = aFlowline[i]
                nEdge = pFlowline.nEdge
                nVertex = pFlowline.nVertex
                aEdge = pFlowline.aEdge

                iStream_segment = pFlowline.iStream_segment
                iStream_order = pFlowline.iStream_order

                for j in range(nEdge):
                    pEdge = aEdge[j]
                    pVertex_start = pEdge.pVertex_start
                    pVertex_end = pEdge.pVertex_end
                    for k in range(ncell):
                        pVertex_center = aCell_all[k].pVertex_center

                        if pVertex_center == pVertex_start:
                            aCell_all[k].iStream_segment_burned = iStream_segment
                            aCell_all[k].iStream_order_burned = iStream_order

                            for l in range(ncell):
                                pVertex_center2 = aCell_all[l].pVertex_center
                                lCellID = aCell_all[l].lCellID
                                if pVertex_center2 == pVertex_end:
                                    aCell_all[k].lCellID_downstream_burned = lCellID
                                    if lCellID in aCellID_outlet:
                                        aCell_all[l].iStream_segment_burned = iStream_segment
                                        aCell_all[l].iStream_order_burned = iStream_order

                                    break
                                
                                

                pass
        else:
            #only mesh, no flowline
            pass

        with open(sFilename_json, 'w', encoding='utf-8') as f:
            sJson = json.dumps([json.loads(ob.tojson()) for ob in aCell_all], indent = 4)        
            f.write(sJson)    
            f.close()

        return

    def tojson(self):  
        aSkip = ['aBasin', \
                'aFlowline_simplified','aFlowline_conceptual','aCellID_outlet',
                'aCell']      

        obj = self.__dict__.copy()
        for sKey in aSkip:
            obj.pop(sKey, None)
        sJson = json.dumps(obj,\
            sort_keys=True, \
                indent = 4, \
                    ensure_ascii=True, \
                        cls=CaseClassEncoder)
        return sJson

    def export_config_to_json(self, sFilename_output):
        aSkip = ['aBasin', \
                'aFlowline_simplified','aFlowline_conceptual','aCellID_outlet',
                'aCell']
        obj = self.__dict__.copy()
        for sKey in aSkip:
            obj.pop(sKey, None)
        with open(sFilename_output, 'w', encoding='utf-8') as f:
            json.dump(obj, f,sort_keys=True, \
                ensure_ascii=False, \
                indent=4, \
                cls=CaseClassEncoder)
        return




import numpy as np
import json

def export_mesh_info_to_json(iFlag_flowline_in, aCell_in, aFlowline_in, aCellID_outlet_iin, sFilename_json_in):
    ncell=len(aCell_in)

    if iFlag_flowline_in == 1:
        nFlowline = len(aFlowline_in)
        for i in range(nFlowline):
            pFlowline = aFlowline_in[i]
            nEdge = pFlowline.nEdge
            nVertex = pFlowline.nVertex
            aEdge = pFlowline.aEdge

            iStream_segment = pFlowline.iStream_segment
            iStream_order = pFlowline.iStream_order

            for j in range(nEdge):
                pEdge = aEdge[j]
                pVertex_start = pEdge.pVertex_start
                pVertex_end = pEdge.pVertex_end
                for k in range(ncell):
                    pVertex_center = aCell_in[k].pVertex_center

                    if pVertex_center == pVertex_start:
                        aCell_in[k].iStream_segment_burned = iStream_segment
                        aCell_in[k].iStream_order_burned = iStream_order

                        for l in range(ncell):
                            pVertex_center2 = aCell_in[l].pVertex_center
                            lCellID = aCell_in[l].lCellID
                            if pVertex_center2 == pVertex_end:
                                aCell_in[k].lCellID_downstream_burned = lCellID
                                if lCellID in aCellID_outlet_iin:
                                    aCell_in[l].iStream_segment_burned = iStream_segment
                                    aCell_in[l].iStream_order_burned = iStream_order

                                break
                            
                            

            pass
    else:
        #only mesh, no flowline
        pass



    with open(sFilename_json_in, 'w', encoding='utf-8') as f:
        sJson = json.dumps([json.loads(ob.tojson()) for ob in aCell_in], indent = 4)        
        f.write(sJson)    
        f.close()
    
    return




from pyflowline.algorithms.auxiliary.find_index_in_list import add_unique_vertex

def find_flowline_vertex(aFlowline_in):
   
    
    nFlowline = len(aFlowline_in)
 
    aVertex=list()
    for i in range(0, nFlowline):      
        pFlowline = aFlowline_in[i]
        pVertex_start = pFlowline.pVertex_start
        pVertex_end = pFlowline.pVertex_end
        aVertex, dummy = add_unique_vertex(aVertex, pVertex_start)
        aVertex, dummy = add_unique_vertex(aVertex, pVertex_end)
        pass


    aVertex_middle=list()
    for i in range(0, nFlowline):      
        pFlowline = aFlowline_in[i]
        aVertex2 = pFlowline.aVertex
        nVertex = pFlowline.nVertex
        for j in range(1, nVertex-1):   
            pVertex= aVertex2[j]
            aVertex_middle.append( pVertex)

    #aVertex_middle2 = list()
    #for elem in aVertex_middle:
    #    if aVertex_middle.count(elem) > 1:
    #        aVertex_middle2, dummy = add_unique_vertex(aVertex_middle2, elem)            
    #    else:
    #        pass

    #aVertex_middle2 =  [x for n, x in enumerate(aVertex_middle) if x in aVertex_middle[:n]] #[item for item, count in collections.Counter(aVertex_middle).items() if count > 1]
    
    #merge
    #aVertex = list(set().union(aVertex, aVertex_middle2))


    return aVertex



import os
import numpy as np
from osgeo import ogr, osr
from shapely.ops import polygonize
from pyflowline.algorithms.auxiliary.gdal_functions import  calculate_angle_betwen_vertex_normal
from pyflowline.algorithms.auxiliary.gdal_functions import calculate_polygon_area
from pyflowline.algorithms.auxiliary.find_index_in_list import find_list_in_list ,find_vertex_in_list

def calculate_area_of_difference_raw(sFilename_a, sFilename_b):
    #not yet supported
    return

def calculate_area_of_difference_simplified(aFlowline_in, aVertex_all_in, \
     sFilename_output_in):

    if os.path.exists(sFilename_output_in): 
        #delete it if it exists
        os.remove(sFilename_output_in)
        pass

    pDriver_geojson = ogr.GetDriverByName( "GeoJSON")

    nFlowline = len(aFlowline_in)
    nVeretx = len(aVertex_all_in)

    #rebuild
    for i  in range(nFlowline):
        aFlowline_in[i].lFlowlineID = i
    for i  in range(nVeretx):
        aVertex_all_in[i].lVertexID = i

    for i  in range(nVeretx):
    #for n, v in enumerate(aVertex_all_in):
    
        for j in range(nFlowline):
            if aVertex_all_in[i] == aFlowline_in[j].pVertex_start:
                aFlowline_in[j].pVertex_start.lVertexID = i

            elif aVertex_all_in[i] == aFlowline_in[j].pVertex_end:
                aFlowline_in[j].pVertex_end.lVertexID = i

    for pFlowline in aFlowline_in:      
        for j  in range(len(aFlowline_in)):
            pFlowline2 = aFlowline_in[j]
            if pFlowline.lFlowlineID != pFlowline2.lFlowlineID:
                if (pFlowline.pVertex_start == pFlowline2.pVertex_start ):
                    pFlowline.aFlowlineID_start_start.append(pFlowline2.lFlowlineID)

                if (pFlowline.pVertex_start==pFlowline2.pVertex_end):
                    pFlowline.aFlowlineID_start_end.append(pFlowline2.lFlowlineID)   
                
                if (pFlowline.pVertex_end==pFlowline2.pVertex_start ):
                    pFlowline.aFlowlineID_end_start.append(pFlowline2.lFlowlineID)
                if (pFlowline.pVertex_end==pFlowline2.pVertex_end):
                    pFlowline.aFlowlineID_end_end.append(pFlowline2.lFlowlineID)

    def get_next_branch(iFlag_rightleft, iFlag_reverse, pFlowline_in):
        """
        This function starts at a contact (called `edge_0`) and calculates
        the angles between the adjacent contacts at a side (either `stop` or
        `start`) and then chooses the right-branching contact, and the opposite
        side for the next contact.

        For example, if we are at `edge_0` and `edge_2` is the right-branching contact,
        and the `start` side of `edge_2` meets `edge_0`, then the next side will be `stop`.

        """

        nEdge = pFlowline_in.nEdge
        iFlag_exist = 0 
        if iFlag_reverse ==0 :
            #normal direction
            pVertex_stop = pFlowline_in.pVertex_end
            pEdge = pFlowline_in.aEdge[nEdge-1]

            x1 = pEdge.pVertex_start.dLongitude_degree
            y1 = pEdge.pVertex_start.dLatitude_degree
            x2 = pEdge.pVertex_end.dLongitude_degree
            y2 = pEdge.pVertex_end.dLatitude_degree
            
           
            iIndex_right = -1

            n = len(pFlowline_in.aFlowlineID_end_start) + len(pFlowline_in.aFlowlineID_end_end)
            if n == 0:
                iFlag_exist = 0 
                iFlag_reverse_new= 0
                pFlowline_out = None
                pVertex_stop_out = None
                return iFlag_exist, iFlag_reverse_new, pFlowline_out, pVertex_stop_out
            else:
                aAngle=list()
                for i in pFlowline_in.aFlowlineID_end_start:
                    pFlowline_dummy = aFlowline_in[i]
                    if (pFlowline_dummy.pVertex_start == pVertex_stop):
                        nEdge2 = pFlowline_dummy.nEdge
                        pEdge2 = pFlowline_dummy.aEdge[0]
                        pVertex_dummy = pEdge2.pVertex_end
                    else:                        #revered
                        nEdge2 = pFlowline_dummy.nEdge
                        pEdge2 = pFlowline_dummy.aEdge[nEdge2-1]
                        pVertex_dummy = pEdge2.pVertex_start

                    #calculate angle              
                    x3 = pVertex_dummy.dLongitude_degree
                    y3 = pVertex_dummy.dLatitude_degree
                    angle_dummy = calculate_angle_betwen_vertex_normal( x1, y1, x2, y2, x3, y3  )                    
                    aAngle.append(angle_dummy)
                for i in pFlowline_in.aFlowlineID_end_end:
                    pFlowline_dummy = aFlowline_in[i]
                    if (pFlowline_dummy.pVertex_start == pVertex_stop):
                        nEdge2 = pFlowline_dummy.nEdge
                        pEdge2 = pFlowline_dummy.aEdge[0]
                        pVertex_dummy = pEdge2.pVertex_end
                    else:                        #revered
                        nEdge2 = pFlowline_dummy.nEdge
                        pEdge2 = pFlowline_dummy.aEdge[nEdge2-1]
                        pVertex_dummy = pEdge2.pVertex_start

                    #calculate angle              
                    x3 = pVertex_dummy.dLongitude_degree
                    y3 = pVertex_dummy.dLatitude_degree
                    angle_dummy = calculate_angle_betwen_vertex_normal( x1, y1, x2, y2, x3, y3  )                    
                    aAngle.append(angle_dummy)
                    

                #mini
                dummy = pFlowline_in.aFlowlineID_end_start + pFlowline_in.aFlowlineID_end_end
                if iFlag_rightleft ==0 :
                    min_angle = min(aAngle)
                    iIndex_right = aAngle.index(min_angle)                   
                    pFlowline_out = aFlowline_in[ dummy[iIndex_right] ]
                    pass
                else:
                    max_angle = max(aAngle)
                    iIndex_left = aAngle.index(max_angle)
                    pFlowline_out = aFlowline_in[ dummy[iIndex_left]]
                    pass

                if pFlowline_out.pVertex_start == pVertex_stop:                 
                    iFlag_reverse_new=0
                    pVertex_stop_out =   pFlowline_out.pVertex_end
                else:                  
                    iFlag_reverse_new=1
                    pVertex_stop_out=pFlowline_out.pVertex_start

                iFlag_exist = 1
                return iFlag_exist, iFlag_reverse_new, pFlowline_out, pVertex_stop_out


        else:
            #rever direction
            pVertex_stop = pFlowline_in.pVertex_start
            pEdge = pFlowline_in.aEdge[0]            
            x1 = pEdge.pVertex_end.dLongitude_degree
            y1 = pEdge.pVertex_end.dLatitude_degree
            x2 = pEdge.pVertex_start.dLongitude_degree
            y2 = pEdge.pVertex_start.dLatitude_degree
            angle_min = 360
            iIndex_right = -1
            n = len(pFlowline_in.aFlowlineID_start_start) + len(pFlowline_in.aFlowlineID_start_end)
            if n == 0:
                iFlag_exist = 0 
                iFlag_reverse_new= 0
                pFlowline_out = None
                pVertex_stop_out = None
                return iFlag_exist, iFlag_reverse_new, pFlowline_out, pVertex_stop_out
            else:
                aAngle=list()
                for i in pFlowline_in.aFlowlineID_start_start:
                    pFlowline_dummy = aFlowline_in[i]
                    if (pFlowline_dummy.pVertex_start == pVertex_stop):
                        nEdge2 = pFlowline_dummy.nEdge
                        pEdge2 = pFlowline_dummy.aEdge[0]
                        pVertex_dummy = pEdge2.pVertex_end
                    else:
                        nEdge2 = pFlowline_dummy.nEdge
                        pEdge2 = pFlowline_dummy.aEdge[nEdge2-1]
                        pVertex_dummy = pEdge2.pVertex_start

                    x3 = pVertex_dummy.dLongitude_degree
                    y3 = pVertex_dummy.dLatitude_degree
                    angle_dummy = calculate_angle_betwen_vertex_normal( x1, y1, x2, y2, x3, y3  )
                    aAngle.append(angle_dummy)
                for i in pFlowline_in.aFlowlineID_start_end:
                    pFlowline_dummy = aFlowline_in[i]
                    if (pFlowline_dummy.pVertex_start == pVertex_stop):
                        nEdge2 = pFlowline_dummy.nEdge
                        pEdge2 = pFlowline_dummy.aEdge[0]
                        pVertex_dummy = pEdge2.pVertex_end
                    else:
                        nEdge2 = pFlowline_dummy.nEdge
                        pEdge2 = pFlowline_dummy.aEdge[nEdge2-1]
                        pVertex_dummy = pEdge2.pVertex_start

                    x3 = pVertex_dummy.dLongitude_degree
                    y3 = pVertex_dummy.dLatitude_degree
                    angle_dummy = calculate_angle_betwen_vertex_normal( x1, y1, x2, y2, x3, y3  )
                    aAngle.append(angle_dummy)
                    
                #mini
                dummy = pFlowline_in.aFlowlineID_start_start + pFlowline_in.aFlowlineID_start_end
                if iFlag_rightleft ==0 :
                    min_angle = min(aAngle)
                    iIndex_right = aAngle.index(min_angle)
                    pFlowline_out = aFlowline_in[dummy[iIndex_right]]
                    pass
                else:
                    max_angle = max(aAngle)
                    iIndex_left = aAngle.index(max_angle)
                    pFlowline_out = aFlowline_in[dummy[iIndex_left]]
                    pass

                if pFlowline_out.pVertex_start == pVertex_stop:            
                    iFlag_reverse_new=0
                    pVertex_stop_out =   pFlowline_out.pVertex_end
                else:           
                    iFlag_reverse_new=1
                    pVertex_stop_out=pFlowline_out.pVertex_start

                iFlag_exist = 1
                return iFlag_exist, iFlag_reverse_new, pFlowline_out, pVertex_stop_out
  

    def walk_cycle(iFlag_rightleft, iFlag_reverse, pFlowline_in, pVertex_origin, aFlowline_list, aVertex_list ):
        iFlag_loop = 1
        if iFlag_reverse ==0:

            iFlag_right = pFlowline_in.iFlag_right
            if iFlag_right == 1:
                iFlag_loop = 0
                return iFlag_loop, aFlowline_list
            else:
                
                iFlag_exist, iFlag_reverse_new, pFlowline_next, pVertex_stop  = get_next_branch(iFlag_rightleft, 0, pFlowline_in)
                if iFlag_exist ==1:
                    aFlowline_list.append(pFlowline_next.lFlowlineID)
                    if (pVertex_stop == pVertex_origin):
                        #a loop is finished
                        iFlag_loop = 1
                        return iFlag_loop, aFlowline_list
                        pass
                    else:
                        iFlag_vertex_exist, index_dummy = find_vertex_in_list(aVertex_list, pVertex_stop)
                        if iFlag_vertex_exist ==1:
                            #this one show before
                            iFlag_loop = 0
                            return iFlag_loop, aFlowline_list
                        else:
                            aVertex_list.append(pVertex_stop)

                        pFlowline_in = pFlowline_next
                        iFlag_loop, aFlowline_list = walk_cycle(iFlag_rightleft, iFlag_reverse_new, pFlowline_in,\
                            pVertex_origin, aFlowline_list, aVertex_list)
                else:
                    iFlag_loop = 0
                    return iFlag_loop, aFlowline_list

        else:
            iFlag_right = pFlowline_in.iFlag_right
            if iFlag_right == 1:
                iFlag_loop = 0
                return iFlag_loop, aFlowline_list
            else:
                
                iFlag_exist, iFlag_reverse_new, pFlowline_next, pVertex_stop = get_next_branch(iFlag_rightleft,1, pFlowline_in)
                
                if iFlag_exist ==1:
                    #aFlowline_in[pFlowline_in.lFlowlineID].iFlag_right=1

                    aFlowline_list.append(pFlowline_next.lFlowlineID)

                    if (pVertex_stop == pVertex_origin):
                        #a loop is finished
                        iFlag_loop = 1
                        return iFlag_loop, aFlowline_list                        
                    else:
                        iFlag_vertex_exist, index_dummy = find_vertex_in_list(aVertex_list, pVertex_stop)
                        if iFlag_vertex_exist ==1:
                            #this one show before
                            iFlag_loop = 0
                            return iFlag_loop, aFlowline_list
                        else:
                            aVertex_list.append(pVertex_stop)

                        pFlowline_in = pFlowline_next
                        iFlag_loop, aFlowline_list = walk_cycle(iFlag_rightleft, iFlag_reverse_new, pFlowline_in, \
                        pVertex_origin, aFlowline_list,aVertex_list)
                else:
                    iFlag_loop = 0
                    return iFlag_loop, aFlowline_list

                
        return iFlag_loop, aFlowline_list

    aList_all = list()

    for i in range(0,nFlowline):
        pFlowline_in = aFlowline_in[i]
        aFlowline_list=list()
        aVertex_list=list()
        iFlag_loop , aFlowline_list = walk_cycle(0, 0, pFlowline_in, pFlowline_in.pVertex_start, aFlowline_list, aVertex_list)
        if iFlag_loop == 1:            
            aFlowline_list.insert(0, i)
            iFlag = find_list_in_list(aList_all,aFlowline_list )
            if iFlag == 1:
                pass
            else:
                aList_all.append(aFlowline_list)
                

    
    pDataset = pDriver_geojson.CreateDataSource(sFilename_output_in) 
    pSpatial_reference_gcs = osr.SpatialReference()  
    pSpatial_reference_gcs.ImportFromEPSG(4326)    # WGS84 lat/lon
    pSpatial_reference_gcs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

    pLayer = pDataset.CreateLayer('aod', pSpatial_reference_gcs, ogr.wkbPolygon)
    # Add one attribute
    pLayer.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution

    pLayerDefn = pLayer.GetLayerDefn()
    pFeature = ogr.Feature(pLayerDefn)    
    lCellID = 0
    dArea = 0.0

    #shapely method, but not complete
    for n in range(0,len(aList_all)):
        aFlowline_list = aList_all[n]
        nFlowline = len(aFlowline_list)
        aFlowline_polygon=list()
        for m in range(nFlowline):
            dummy_index = aFlowline_list[m]
            pFlowline = aFlowline_in[dummy_index]
            nVertex= pFlowline.nVertex
            aCoords_gcs = np.full((nVertex,2), -9999. ,dtype=float)
            for k in range(nVertex):
                aCoords_gcs[k,0] = pFlowline.aVertex[k].dLongitude_degree
                aCoords_gcs[k,1] = pFlowline.aVertex[k].dLatitude_degree
            
            aCoords_gcs= tuple(j for j in aCoords_gcs) #np.array(aCoords_gcs)    
            aFlowline_polygon.append(aCoords_gcs)   

        dummy = polygonize(aFlowline_polygon)
        aPolygon_out =  list(dummy) 
  
        for po in aPolygon_out:
            ring = ogr.Geometry(ogr.wkbLinearRing)
            aCoords_gcs = po.exterior.coords
            aCoords_gcs= np.array(aCoords_gcs)  
            nPoint  = (aCoords_gcs.shape)[0]
            lons=list()
            lats=list()
            for i in range(nPoint):
                ring.AddPoint(aCoords_gcs[i,0], aCoords_gcs[i,1])
                lons.append( aCoords_gcs[i,0] )
                lats.append( aCoords_gcs[i,1] )

            pPolygon = ogr.Geometry(ogr.wkbPolygon)
            pPolygon.AddGeometry(ring)
            pFeature.SetGeometry(pPolygon)
            pFeature.SetField("id", lCellID)
            pLayer.CreateFeature(pFeature)
            lCellID= lCellID+1

            dArea0 = calculate_polygon_area(lons, lats)
            dArea = dArea + dArea0

    return aPolygon_out, dArea
   
    

    

      

             

    
